import arsd.dom;
import arsd.cgi;

import std.array;
import std.string;
import std.conv;

// I'm using this as a delimiter string because dom.d will never output it - it will always be encoded as an entity in the main section, and should be rare enough in RawSource segments not to worry too much about
enum dStringDelimiterOpen = `cgi.write(q"DHP_LITERAL` ~ "\n";
enum dStringDelimiterClose = "\n" ~ `DHP_LITERAL"[0 .. $-1]);`;

string getDCode(string dhpCode) {
	auto document = new Document();
	document.enableAddingSpecialTagsToDom();
	document.parse(dhpCode, true, true);

	auto w = appender!string();
	w.put(`import arsd.cgi;
pragma(mangle, "handler") void handler(Cgi cgi) {
`);
	w.put(dStringDelimiterOpen);

	class DCode : AspCode {
		this(Document p, string t) { super(p, t); }
		override string writeToAppender(Appender!string where = appender!string()) const {
			auto start = where.data.length;
			where.put(dStringDelimiterClose);
			where.put(this.source);
			where.put(dStringDelimiterOpen);
			return where.data[start .. $];
		}
	}

	foreach(element; document.root.tree) {
		if(auto asp = cast(AspCode) element) {
			auto dc = new DCode(asp.parentDocument, "d");
			dc.source = asp.source[1 .. $-1];
			asp.replaceWith(dc);
		}
	}

	document.root.writeToAppender(w);

	w.put(dStringDelimiterClose);
	w.put("\n}");

	// get rid of empty writes since those might mess up caching
	return w.data.replace(dStringDelimiterOpen ~ dStringDelimiterClose, "");
}

extern(C) void *dlopen(in char *name, int flags);
extern(C) void *dlsym(void *handle, const char *symbol);
extern(C) int dlclose(void *handle);
static __gshared void*[string] libraries;

void handleRequest(Cgi cgi) {
	int pathInfoStart = 0;
	string soName = cgi.pathInfo;

	assert(soName.length);
	if(soName.length == 1) {
		soName = "index";
	} else {
		soName = soName[1 .. $];
		pathInfoStart = 1;
		auto slashIdx = soName.indexOf("/");
		if(slashIdx != -1) {
			soName = soName[0 .. slashIdx];
			pathInfoStart += slashIdx;
		}
	}

	{
		import std.regex;

		// trim whatever so we don't try to traverse the filesystem, restrictive
		// name should keep us right here in this folder with no surprises
		// also makes it safe to just drop right into a call to system()
		soName = std.regex.replace(soName, regex(`[^a-z0-9\-]`, "g"), "");
	}

	void* library;

	synchronized {
		auto libResult = soName in libraries;
		if(libResult !is null)
			library = *libResult;
	}


	import std.file;
	import std.process;
	try {
		if(!exists(soName ~ ".so") || timeLastModified(soName ~ ".dhp") > timeLastModified(soName ~ ".so")) {
		synchronized {
			if(library !is null)
				dlclose(library);
			library = null;
				libraries.remove(soName);

			auto dhpCode = readText(soName ~ ".dhp");
			auto dCode = getDCode(dhpCode);
			std.file.write(soName ~ ".d", dCode);
			auto dmdOutput = executeShell("dmd -shared "~soName~".d arsd/cgi.d");
			if(dmdOutput.status != 0) {
				string[] errors;
				auto codeLines = dCode.splitLines();
				auto originalLines = dhpCode.splitLines();
				int lastBadLine = 0;
				foreach(line; dmdOutput.output.splitLines) {
					auto colon = line.indexOf(":");
					if(colon != -1) {
						auto fileAndLine = line[0 .. colon];
						auto message = line[colon .. $];

						auto paren = fileAndLine.indexOf("(");
						if(paren != -1) {
							auto file = fileAndLine[0 .. paren];
							auto lineNumber = to!int(fileAndLine[paren + 1 .. $ -1]); // $-1 cuts off the right hand )

							if(file == soName ~ ".d") {
								// take out some trash
								message = message.replace(dStringDelimiterClose, "");
								message = message.replace(dStringDelimiterOpen, "");

								// fix up the line number and filename
								file = file.replace(".d", ".dhp");
								auto complaintLine = dCode[lineNumber - 1];
								// the original line will always be ahead of this one because we add lines when building the D
								foreach(index, possibleLine; originalLines[lastBadLine .. lineNumber > originalLines.length ? originalLines.length : lineNumber]) {
									// probably it
									if(possibleLine.indexOf(complaintLine) != -1) {
										lineNumber = index + lastBadLine;
										lastBadLine = lineNumber;
										lineNumber++; // lines start at 1, not 0
										break;
									}
								}

								// and record the error
								errors ~= format("%s(%s)%s", file, lineNumber, message);
							}
							// otherwise, it is irrelevant (or is it??????)
						}
					}
				}

				cgi.setResponseStatus("500 Compilation failed");
				cgi.setResponseContentType("text/plain");
				if(errors.length)
					cgi.write(errors.join("\n"));
				else // if we didn't find any errors, something still went wrong, so this way the user isn't completely in the dark
					cgi.write(dmdOutput.output);
				cgi.close();
				return;
			}
		}
		}
	} catch(Exception e) {
		cgi.setResponseStatus("500 Internal Server Error");
		cgi.setResponseContentType("text/plain");
		cgi.write(e.toString());
		cgi.close();
		return;
	}

	if(library is null) {
		library = dlopen(toStringz("./" ~ soName ~ ".so"), 2 /* RTLD_NOW */);

		if(library is null)
			throw new Exception("couldn't open the .so");

		synchronized
			libraries[soName] = library;
	}

	auto symbol = dlsym(library, "handler");
	if(symbol is null) {
		throw new Exception("invalid .so");
	}

	auto handler = cast(void function(Cgi)) symbol;

	// kids, don't do this at home
	// cheating to adjust the cgi object to reflect our new location
	cast(string)(cgi.scriptName) = soName ~ ".dhp";
	cast(string)(cgi.pathInfo) = cgi.pathInfo[pathInfoStart .. $];

	handler(cgi);
}

mixin GenericMain!handleRequest;
